/*
JavaScript tools shared between Fluid web views.
*/

class DummyPendingPromise {
  resolve(value) {}
  reject(message) {}
}

class PendingPromise extends DummyPendingPromise {
  constructor(resolve, reject) {
    super();
    this._resolve = resolve;
    this._reject = reject;
  }

  resolve(value) {
    super.resolve(value);
    this._resolve(value);
  }

  reject(message) {
    super.reject(message);
    this._reject(new Error(message));
  }
}

// Registry of pending promises that can be resolved externally.
// One instance of the registry exists in the fluidTools.promises variable.
class PromiseRegistry {
  constructor() {
    this._pending = {};
    this._next = 0;
  }

  reset() {
    // Clear all pending promises, but do not reset the counter so a delayed call from the host
    // won't accidentally reset or reject a new promise.
    this._pending = {};
  }

  register(action) {
    const self = this;

    return new Promise(function(resolve, reject) {
      try {
        const id = self._next;
        action(id);
        self._next += 1;
        self._pending[id] = new PendingPromise(resolve, reject);
      } catch (ex) {
        console.error("PromiseRegistry.register|".concat(ex.toString()));
        reject(ex);
      }
    });
  }

  resolve(id, value) {
    console.log("resolve|id=".concat(id));
    this.pop(id).resolve(value);
  }

  reject(id, message) {
    console.error("reject|id=".concat(id, "|message=", message));
    this.pop(id).reject(message);
  }

  pop(id) {
    let promise = this._pending[id];

    if (promise != null) {
        delete this._pending[id];
    } else {
        promise = new DummyPendingPromise();
    }

    return promise;
  }
}

class FluidTokenProvider {

  // getToken - function that the token provider calls to fetch a token; must take 1 parameter:
  //              scopes - array of scope resource strings for the token
  constructor(getToken) {
    this._getToken = getToken;
  }

  getToken(scopes) {
    const self = this;
    const action = function(id) {
      // Call the hosting app to fetch the requested token and resolve the promise with the allocated id.
      // To resolve the promise the host must execute the following script in the web view:
      //    fluidTools.promises.resolve(<integer ID>, "<obtained token>");
      // To reject the promise the host executes this:
      //    fluidTools.promises.reject(<integer ID>, "<error message>");
      if (!self._getToken(id, scopes)) {
        const e = new Error("Failed to request a token|".concat(id));
        e.name = "FrameworkError";
        throw e;
      }
    };

    return fluidTools.promises.register(action);
  }
}

class FluidFailureProcessor {

  constructor() {
    this._redeemCount = 0;
  }

  reset() {
    this._redeemCount = 0;
  }

  //
  // Called by the script in the layout when the promise returned by the framework loader rejects with an error object.
  // Returns a promise that resolves if the error can be fixed by redeeming a shared link to the Fluid document that has failed to load
  // and calling the loader again.
  // If redeeming cannot fix the error, or the maximum number of attempts to redeem has been made, the returned promise rejects with the original error.
  //
  canRedeem(error) {

    const REDEEM_LIMIT = 1;

    const self = this;
    const sError = fluidTools.stringify(error);

    return new Promise((resolve, reject) => {
      try {
        console.log("canRedeem|".concat(sError));

        if (this.isEligibleForRedeeming(error)) {
          if (self._redeemCount < REDEEM_LIMIT) {
            console.log("Eligible for redeeming|".concat(sError));
            self._redeemedCount += 1;
            resolve();
          } else {
            reject(error);
          }
        } else {
          reject(error);
        }
      } catch (ex) {
        console.error("canRedeem|unexpected exception|".concat(self.stringify(ex)));
        reject(ex);
      }
    });
  }

  isEligibleForRedeeming(error) {
    let eligible = false;

    // Go through the tree of nesting errors error -> error.internalError and check
    // if the error name at some level is "FileNotFound", which may indicate that the framework got an HTTP 404
    // from SharePoint.
    // Stop at a "RuntimeError" error because it indicates an exception thrown in JavaScript, which is irrecoverable.
    for (let e = error; !eligible && e != null && e.name != "RuntimeError"; e = e.internalError) {
    // TODO this error name can change which will break redeem logic, we need a better way to control the flow
    // for redeem
      if (e.name.startsWith("FileNotFound")) {
        eligible = true;
      }
    }

    return eligible;
  }
}

const ComponentType = {
  TABLERO: 1,
  SCRIPTOR: 2,
  EMPTY: 3,
  UNKNOWN: 4
};

//
// OWH/Fluid framework wrapper that the script in the HTML layout calls to :
//  - Load a Fluid document
//  - Compose a Fluid document
//
class FluidClientTools {
  constructor(framework) {
    // The host replaces %hostThemeData% with a JSON string. If the string will fail to parse, there will be no theming.
    const injectedHostThemeData = '%hostThemeData%';
    const injectedInitialTheme = '%hostInitialTheme%';

    const fixUp = function(injectedValue, fallbackValue) {

        let fixedUp = injectedValue;

        if (fixedUp != null) {
            fixedUp = fixedUp.toString();

            if (fixedUp.length < 2) {
                console.log("FluidClientTools.fixUp|not a string or too short|".concat(injectedValue));
                fixedUp = null;
            } else if (fixedUp.startsWith('%') && fixedUp.endsWith('%')) {
                console.log("FluidClientTools.fixUp|not injected|".concat(injectedValue));
                fixedUp = null;
            }
        }

        if (fixedUp == null) {
            fixedUp = fallbackValue;
        }

        return fixedUp;
    };

    this.framework = framework;
    this.logger = {};
    this.promises = new PromiseRegistry();
    this.failureProcessor = new FluidFailureProcessor();
    this.hostInitialTheme = injectedInitialTheme;

    try {
      // Wrap this code in the exception handler to get some protection against incorrectly injected host variables.
      this.hostContext = {
          // export type Audience = 'Test' | 'Local' | 'Staging' | 'Dogfood' | 'MSIT' | 'FirstRelease' | 'Production';
          audience: fixUp('%hostAudience%', "Local"),

          hostCorrelationId: fixUp('%hostCorrelationId%', "uncorrelated"),

          hostName: fixUp('%hostHostName%', "android-device"),

          hostAppVersion: fixUp("%hostAppVersion%", "unknown"),

          userContext: {
              user: { oid: fixUp('%hostUserOid%', "no-user"),
              tenantId: fixUp('%hostTenantId%', "no-tenant") }
          },
          hostScenarioName: fixUp('%hostScenarioName%', "no-scenario")
      };


      try {
          this.hostContext.hostScenarioStartTime = Number(fixUp('%hostScenarioStartTime%', Date.now().toString()));
          console.log("FluidClientTools|hostScenarioStartTime=".concat(this.hostContext.hostScenarioStartTime));
      } catch (ex) {
          console.error("FluidClientTools|set scenario start time|".concat(ex.toString()));
          this.hostContext.hostScenarioStartTime = Date.now();
      }


    } catch (ex) {
      console.error("FluidClientTools|build host context|".concat(ex.toString()));
      this.hostContext = {
          audience: "Local",
          hostCorrelationId: "uncorrelated",
          hostName: "android-device",
          hostAppVersion: "unknown",
          userContext: { user: undefined },
          hostScenarioName: "no-scenario",
          hostScenarioStartTime: Date.now()
      };
    }

    console.log("FluidClientTools|hostContext=".concat(this.stringify(this.hostContext)));

    try {
      this.themeControl = framework.makeStylingProvider(JSON.parse(injectedHostThemeData), injectedInitialTheme);
    } catch (ex) {
      console.error("FluidClientTools|Failed to parse the host theme data|".concat(ex.toString(), "|", injectedHostThemeData));
      this.themeControl = null;
    }
  }

  // Reset the internal state of the wrapper to the initial state - clear all registered pending promises
  // and reset the failure processor.
  // The method must be called if the host wants to reload the Fluid document without reloading of the HTML layout.
  reset() {
    this.promises.reset();
    this.failureProcessor.reset();
  }

  makeResizeObserver(callback) {
    return this.framework.makeResizeObserver(callback);
  }

  makeEventDebouncer(delay, valueProvider) {
    return this.framework.makeEventDebouncer(delay, valueProvider);
  }

  // Apply a named theme to the Fluid content.
  // Themes come from the object injected in the constructor (injectedHostThemeData above) that must have the following structure:
  //  {
  //    <theme name 1>: {
  //      themeProperties: {
  //        "property name": "property value",
  //        ...
  //      },
  //      textStyles: {
  //        TBD
  //      }
  //    },
  //    <theme name 2>: {
  //    }
  //    ...
  //  }
  applyTheme(themeName) {
    if (this.themeControl != null) {
        this.themeControl.applyTheme(themeName);
    }
  }

  injectStylingProvider(scopes) {
      const frameworkScopes = scopes == null ? {} : scopes;

      if (this.themeControl != null) {
        console.log("load.onBootstrapped|applying theme|".concat(this.hostInitialTheme));

        const self = this;

        const applyThemeBackground = function(oldThemeProperties) {
            let bgColor = null;

            try {
              bgColor = self.themeControl.themeProperties.get(self.framework.ThemeColorPropertyNames.BodyBackground)
            } catch (ex) {
              console.error("load.onBootstrapped.applyThemeBackground|".concat(ex.toString()));
            }

            if (bgColor == null) {
              bgColor = "#fff";
            }

            console.log("load.onBootstrapped.applyThemeBackground|document.body.style.background=".concat(bgColor));
            document.body.style.background = bgColor;
        };

        frameworkScopes.StylingProvider = this.themeControl;
        applyThemeBackground();
        this.themeControl.on("themePropertiesChanged", oldProperties => applyThemeBackground());
      }

      return frameworkScopes;
  }

  load(fileUri, rootElement, scopes, hostConfig) {
    const self = this;

    return new Promise((resolve, reject) => {
      try {
        const driverInfo = self.framework.tryResolveDriverInfo(fileUri);
        if (!driverInfo) {
          const e = new Error("No driver info found|".concat(fileUri));
          e.name = "FrameworkError";
          throw e;
        }

        if(Object.entries(self.logger) == 0) {
            self.logger = new self.framework.OfficeFluidLogger({
                persistentProps: { correlationId: self.hostContext.hostCorrelationId, hostScenarioStartTime: self.hostContext.hostScenarioStartTime },
                namespace: 'TeamSpaceHost',
                hostContextProps: self.hostContext
            });
        }

        const additionalProperties = {
            correlationId: self.hostContext.hostCorrelationId,
            hostScenarioStartTime: self.hostContext.hostScenarioStartTime
        };

        const frameworkScopes = self.injectStylingProvider(scopes);
        frameworkScopes.HostContext = self.hostContext;
        hostConfig.logger = self.logger;

        self.framework.loadComponentInDiv(driverInfo, rootElement, frameworkScopes, hostConfig)
          .then(loadedInfo => {
             resolve(loadedInfo); }
          , error => {
             reject(error);
             });
      } catch (ex) {
        console.error("load|".concat(ex.toString()));
        reject(self.wrapException(ex));
      }
    });
  }

   compose(rootElement, scopes, storageInfoPromise, hostConfig) {
        const self = this;

        return new Promise((resolve, reject) => {
          try {
            const newFileInfoPromise = storageInfoPromise.then(storageAndConfig => {
              console.log("onStorageInfoRetrieved: ".concat(storageAndConfig));
              const storageJson = JSON.parse(storageAndConfig);

               return { siteUrl: storageJson.siteUrl,
                        driveId: storageJson.driveId,
                        filename: storageJson.fileName,
                        filePath: storageJson.fileRelativePath };
            });

        const driverInfo = self.framework.resolveODSPDriverInfoForNewFile(newFileInfoPromise);

        if(Object.entries(self.logger) == 0) {
            self.logger = new self.framework.OfficeFluidLogger({
                hostContextProps: self.hostContext,
                persistentProps: {},
                namespace: 'TeamSpaceHost'
            });
        }
        const frameworkScopes = self.injectStylingProvider(scopes);
        frameworkScopes.HostContext = self.hostContext;
        hostConfig.logger = self.logger;

        self.framework.loadComponentInDiv(driverInfo, rootElement, frameworkScopes, hostConfig)
          .then(loadedInfo => {
            resolve(loadedInfo);
          }
            , error => {
              reject(error);
            });
          } catch (exception) {
            console.error("compose|".concat(exception.toString()));
            reject(self.wrapException(exception));
          }
        });
  }

  // limitSize is set via client ECS.
  // If the sizeLimit feature is enabled on client, but the value isn't set up on ECS,  client will take the default value set in native code 1024 * 1024
  // If the sizeLimit feature isn't enabled on client, client will pass 0 to JS
  getHostConfigForLoad(limitSize) {
    return {
        loaderOptions: {},
        componentRenderOptions: { onlyInlineComponents: %onlyInlineComponents% },
        hostStoragePolicy: limitSize > 0 ? {snapshotOptions : {mds : limitSize}} : {}
      };
  }

  getHostConfigForCompose(componentType, componentParams) {
    var containerConfig = {};

    switch (componentType) {
        case ComponentType.TABLERO:
            containerConfig = {
                configurationType: "templateConfiguration",
                loadTemplateData: {
                    configurationId: 'tableInCanvas',
                    defaultComponentInitialConfiguration: {
                        heroComponentPath: 'heroComponentPath'
                    },
                    heroComponentInitialConfiguration: {
                        numOfRows: componentParams.rowCount,
                        columnTypes: Array(componentParams.colCount).fill("IRichTextData")
                    }
                },
            };
            break;
        case ComponentType.SCRIPTOR:
            var listType;
            switch (componentParams.listType) {
                case "number":
                    listType = "numberedList";
                    break;
                case "checked":
                    listType = "checkedList";
                    break;
                case "bullet":
                case "unknown":
                default:
                    listType = 'bulletedList';
                   break;
            }

            containerConfig = {
              configurationType: "templateConfiguration",
              loadTemplateData: {
                configurationId: listType,
                defaultComponentInitialConfiguration: {
                  heroComponentPath: 'heroComponentPath'
                }
              }
            };
            break;
        case ComponentType.EMPTY:
        case ComponentType.UNKNOWN:
        default:
            break;
    }

    return {
      logger: self.logger,
      loaderOptions: {},
      componentRenderOptions: { onlyInlineComponents: %onlyInlineComponents% },
      containerConfiguration: containerConfig
    };
  }

  parseFluidUrl(fluidUrl) {
    const self = this;

    return new Promise((resolve, reject) => {
      try {
        const parsedUrlParams = new URLSearchParams(fluidUrl);
        const fluidParam = parsedUrlParams.get("nav");
        if (!fluidParam) {
          reject({ message: "parseFluidUrl: no fluid query param in URL" });
        } else {
          const urlResolver = new self.framework.OdspUrlResolver();
          urlResolver.resolve({ url: fluidUrl }).then(
            resolved => { resolve({ driveId: resolved.driveId, fileId: resolved.itemId, fluidParam: fluidParam }); },
            error => { reject({ message: error.toString() }); }
          );
        }
      } catch (ex) {
        console.error("parseFluidUrl exception|".concat(self.stringify(ex)));
        reject(ex);
      }
    });
  }

  wrapException(ex) {
    const wrapped = { name: "RuntimeError", internalError: { error: null } };

    if (ex != null) {

      wrapped.type = ex.name;
      wrapped.message = ex.toString();

      for (const prop in ex) {
        const v = ex[prop];
        if (!(v instanceof Function)) {
          wrapped.internalError[prop] = v;
        }
      }
    }
    return wrapped;
  }

  stringify(obj, maxDepth) {
    let str = null;
    const self = this;

    // There is a circular dependency in the error object, copy only string and numerical properties up to 3 levels of the graph
    // to a new object and stringify it.
    const cloneObject = function(original, depth, maxDepth) {
      let clone = {};

      if (original != null && depth < maxDepth) {
        for (const prop in original) {
          const value = original[prop];

          if (value instanceof String || value instanceof Number || value instanceof Date) {
            clone[prop] = value;
          } else if (value instanceof Error) {
            clone[prop] = self.wrapException(value);
          } else if (value instanceof Object) {
            clone[prop] = cloneObject(value, depth + 1, maxDepth);
          }
        }
      }

      return clone;
    };

    if (obj instanceof Error) {
      obj = this.wrapException(obj);
    }

    try {
      str = JSON.stringify(obj);
    } catch (ex) {
      str = JSON.stringify(cloneObject(obj, 0, maxDepth == null ? 3 : maxDepth));
    }

    return str;
  }
}

bootstrapFluidHost({
    tryResolveDriverInfo: null,
    loadComponentInDiv: null,
    resolveODSPDriverInfoForNewFile: null,
    makeStylingProvider: null,
    makeResizeObserver: null,
    makeEventDebouncer: null,
    makeContainerEventFilter: null,
    ThemeColorPropertyNames: null,
    OdspUrlResolver: null,
    OfficeFluidLogger: null
})
.then(
    framework => { window.fluidTools = new FluidClientTools(framework); },
    error => { console.error("bootstrapFluidHost|".concat(error.toString())); });
